//
// F2Window.cpp
//

//
// This software is protected by the GNU General Public License,
// also known as the GPL.
//
// See the file COPYING.txt for details.
//

#include "stdhdr.h"

#include "F2Window.h"

#include "Res\Resource.h"

#include "vgui\ProgressDialog.h"
#include "vgui\BrowseDialog.h"
#include "vgui\Alert.h"
#include "vgui\Question.h"
#include "vgui\Metrics.h"

#include "vos\FileMatches.h"
#include "vos\MonotonicTime.h"
#include "vos\Trace.h"
#include "vos\Preferences.h"

#include "FilenameTransformer.h"
#include "FilenameTransformSpec.h"
#include "FileRenameWorker.h"
#include "OptionsDialog.h"

//
// The message map for this window.
//
// (currently empty)
//

BEGIN_VGUI_MESSAGE_MAP(F2Window, Dialog)
	ON_VGUI_LBN_SELCHANGED(IDC_FILELIST, OnFileSelectionChange)
	ON_VGUI_BN_CLICKED(IDC_BROWSE, OnBrowse)
	ON_VGUI_BN_CLICKED(IDC_RENAME, OnRename)
	ON_VGUI_BN_CLICKED(IDC_UNDO, OnUndo)
	ON_VGUI_BN_CLICKED(IDC_CLEARERRORS, OnClearErrors)
	ON_VGUI_BN_CLICKED(IDC_OPTIONS, OnOptions)
	ON_VGUI_BN_CLICKED(IDC_FILTERFILES, OnFilterFilesClicked)
	ON_VGUI_BN_CLICKED(IDC_FILTER_NORMAL, OnFilterOptionChanged)
	ON_VGUI_BN_CLICKED(IDC_FILTER_REGEXP, OnFilterOptionChanged)
	ON_VGUI_BN_CLICKED(IDC_IGNORECASE, OnFilterOptionChanged)
	ON_VGUI_BN_CLICKED(IDC_SCANFOLDER, OnRefresh)
	ON_VGUI_BN_CLICKED(IDC_SORTONSTATUS, OnSortChange)
	ON_VGUI_BN_CLICKED(IDC_SORTNOW, OnSortNow)
	ON_VGUI_EBN_TEXTCHANGED(IDC_FILTERSPEC, OnFilterOptionChanged)
	ON_VGUI_EBN_TEXTCHANGED(IDC_PATTERN, OnRenameSchemeChanged)
	ON_VGUI_EBN_TEXTCHANGED(IDC_NEWNAME, OnRenameSchemeChanged)
	ON_VGUI_EBN_TEXTCOMMIT(IDC_FOLDERNAME, OnFolderNameChanged)
	ON_VGUI_EBN_TEXTCOMMIT(IDC_FILTERSPEC, OnFilterCommit)
	ON_VGUI_EBN_TEXTCOMMIT(IDC_PATTERN, OnRenameCommit)
	ON_VGUI_EBN_TEXTCOMMIT(IDC_NEWNAME, OnRenameCommit)
END_VGUI_MESSAGE_MAP()

const int kColumn_OriginalName = 0;
const int kColumn_Status = 1;
const int kColumn_NewName = 2;

const vgui::ControlID kFilterGroup[] = 
{
	IDC_FILTERSPEC,
	IDC_FILTER_NORMAL,
	IDC_FILTER_REGEXP,
	IDC_IGNORECASE,
	vgui::NOCTL
};

const vgui::ControlID kProgressGroup[] = 
{
	IDC_LABEL_PROGRESS,
	IDC_PROGRESS,
	vgui::NOCTL
};


const vgui::StringResourceID kResultText[] =
{
	IDS_RENAMESTATE_ERROR_INTERNAL,
	IDS_RENAMESTATE_ERROR_FILENOTFOUND,
	IDS_RENAMESTATE_ERROR_ALREADYEXISTS,
	IDS_RENAMESTATE_ERROR_ACCESSDENIED,
	IDS_RENAMESTATE_ERROR_CREATEFAILED,
	IDS_RENAMESTATE_ERROR_UNKNOWN,

	IDS_RENAMESTATE_WARN_PATTERNFAIL,
	IDS_RENAMESTATE_WARN_NAMECLASH,
	IDS_RENAMESTATE_WARN_BADMETADATA,
	IDS_RENAMESTATE_WARN_FILENOTFOUND,
	IDS_RENAMESTATE_WARN_NEWFILEEXISTS,
	IDS_RENAMESTATE_WARN_READONLY,

	IDS_RENAMESTATE_RENAMESUCCESS,

	IDS_RENAMESTATE_OK,
};

	
//
// Number of milliseconds we delay before showing progress display
// (so we try to avoid annoying flickery progress bars that only appear
// for a split-second).
//
const uint kProgressTimeThreshold = 800;


/**************************************************************************

>	F2Window::F2Window(vgui::Application *pApp)

	Author:		vGUI AppWizard
	Inputs:		pApp - the parent application object.

	Purpose:	Constructor - create the main window object.

**************************************************************************/

F2Window::F2Window(vgui::Application *pApp) : vgui::Dialog(IDD_F2, NULL),
											  m_bProgressDisplayVisible(false),
											  m_bFullProgressDisplayVisible(false),
											  m_TimerID(vgui::NullTimer),
											  m_bDoingBackgroundProcessing(false),
											  m_State(F2State_RenameAllowed),
											  m_bSortByStatusFirst(true)
{
	m_UndoManager.LinkToFilenameList(&m_FilenameList);
}


/**************************************************************************

>	void F2Window::OnInitDialog()

	Author:		Tim Browse
	Date:		28/4/2002

	Purpose:	Attach controls to dialog controls, and initialise
				controls.

**************************************************************************/

void F2Window::OnInitDialog()
{
	// Initialise base class
	Dialog::OnInitDialog();

	m_FolderName.Attach(this, IDC_FOLDERNAME);
	m_FolderName.EnableCommitNotification();
	
	m_UseFilter.Attach(this, IDC_FILTERFILES);
	m_FileFilter.Attach(this, IDC_FILTERSPEC);
	m_FileFilter.EnableCommitNotification();
	m_UseNormalFilter.Attach(this, IDC_FILTER_NORMAL);
	m_UseRegExpFilter.Attach(this, IDC_FILTER_REGEXP);
	m_IgnoreCase.Attach(this, IDC_IGNORECASE);

	m_SortOnStatus.Attach(this, IDC_SORTONSTATUS);
	m_bAutoSizeColumns.Attach(this, IDC_AUTOSIZECOLUMNS);
	m_bAutoSizeColumns = true;

	// Set progress bar to use 0-100 range for percentages
	m_ProgressBar.Attach(this, IDC_PROGRESS);
	m_ProgressBar.SetRange(0,100);
	m_ProgressLabel.Attach(this, IDC_LABEL_PROGRESS);

	vgui::Rect LabelRect;
	GetControlPosition(IDC_LABEL_PROGRESS, &LabelRect);

	vgui::Rect ProgressRect;
	GetControlPosition(IDC_PROGRESS, &ProgressRect);
	m_ProgressWidth = ProgressRect.right - LabelRect.right;

	// Position the error message to cover the progress display controls
	m_ErrorMessage.Attach(this, IDC_ERRORMESSAGE);
	vgui::Rect ErrorMessageRect(LabelRect);
	ErrorMessageRect.right = ProgressRect.right;
	SetControlPosition(IDC_ERRORMESSAGE, ErrorMessageRect);
	HideControl(IDC_ERRORMESSAGE);

	vgui::StringResourceID rErrorMessage(IDS_ERROREXPLANATION);
	m_ErrorMessage.SetText(rErrorMessage);

	m_Pattern.Attach(this, IDC_PATTERN);
	m_Pattern.EnableCommitNotification();
	m_NewName.Attach(this, IDC_NEWNAME);
	m_NewName.EnableCommitNotification();
	m_UseMP3Tags.Attach(this, IDC_USEMETADATA);
	m_FileList.Attach(this, IDC_FILELIST);

	// Set up columns
	m_FileList.SetNumberOfColumns(3);
	m_FileList.SetColumnName(kColumn_Status, "Status");
	m_FileList.SetColumnName(kColumn_OriginalName, "Original Filename");
	m_FileList.SetColumnName(kColumn_NewName, "New Filename");

	// Set up sorting
	m_FileList.SetSortFunction(ListSortFunction, 
							   (vgui::UserData) this, 
							   kColumn_OriginalName);
	m_FileList.SetSortFunction(ListSortFunction, 
							   (vgui::UserData) this, 
							   kColumn_NewName);
	m_FileList.SetSortFunction(ListSortFunction, 
							   (vgui::UserData) this, 
							   kColumn_Status);
	m_bSortByStatusFirst = true;
	m_SortOnStatus = m_bSortByStatusFirst;
	
	// Resize columns appropriately
	m_FileList.AutoSizeColumn(kColumn_Status);
	ResizeFilenameColumns();

#ifdef _DEBUG
	// Enable filter by default
	m_UseFilter = true;
	m_FileFilter = "*.mp3";

	m_FolderName = "C:\\mp3test";
	m_NewName = "%1\\%2\\%3 - %4.mp3";
	m_Pattern = "(.*) - (.*) - (.*) - (.*).mp3";
#else
	// Disable filter by default
	m_UseFilter = false;
	m_FolderName = "C:\\";
	m_NewName = "%1";
	m_Pattern = "(.*)";
#endif
//	m_FolderName = "M:\\library\\AAA - Refiling\\AAA - Refiled\\Ace of Base\\All That She Wants (Single)";
//	m_FolderName = "M:\\library\\AAA - Refiling\\AAA - Orphaned";
//	m_FolderName = "D:\\mp3\\My Music";

//	m_NewName = "%1\\<Album>\\<TrackNumber> - %4.mp3";
//	m_NewName = "%1\\<mp3Album>\\<mp3TrackNumber> - %2.mp3";

//	m_Pattern = "(.*) - (.*).mp3";

	m_UseNormalFilter = true;
	m_IgnoreCase = true;

	if (m_UseFilter)
	{
		EnableControlGroup(kFilterGroup);
	}
	else
	{
		DisableControlGroup(kFilterGroup);
	}
	
	// Initialise progress display visibility state
	HideControlGroup(kProgressGroup);
	m_bProgressDisplayVisible = false;
	m_bFullProgressDisplayVisible = false;

	// Start a folder scan asap
	ForceFolderScan();

	// Set up the controls that change size/position as dialog is resized
	AddDynamicControl(IDC_GROUP_FILES,			vgui::ControlResize_Stretch,	vgui::ControlResize_Stretch);
	AddDynamicControl(IDC_FOLDERNAME,			vgui::ControlResize_Stretch,	vgui::ControlResize_Static);
	AddDynamicControl(IDC_BROWSE,				vgui::ControlResize_Move,		vgui::ControlResize_Static);
	AddDynamicControl(IDC_SCANFOLDER,			vgui::ControlResize_Move,		vgui::ControlResize_Static);

	AddDynamicControl(IDC_FILTERFILES,			vgui::ControlResize_Static,		vgui::ControlResize_Static);
	AddDynamicControl(IDC_FILTERSPEC,			vgui::ControlResize_Stretch,	vgui::ControlResize_Static);
	AddDynamicControl(IDC_FILTER_NORMAL,		vgui::ControlResize_Move,		vgui::ControlResize_Static);
	AddDynamicControl(IDC_FILTER_REGEXP,		vgui::ControlResize_Move,		vgui::ControlResize_Static);
	AddDynamicControl(IDC_IGNORECASE,			vgui::ControlResize_Move,		vgui::ControlResize_Static);

	AddDynamicControl(IDC_SORTONSTATUS,			vgui::ControlResize_Static,		vgui::ControlResize_Move);
	AddDynamicControl(IDC_AUTOSIZECOLUMNS,		vgui::ControlResize_Static,		vgui::ControlResize_Move);
	AddDynamicControl(IDC_FILELIST,				vgui::ControlResize_Stretch,	vgui::ControlResize_Stretch);

	AddDynamicControl(IDC_SORTNOW,				vgui::ControlResize_Move,		vgui::ControlResize_Move);

	AddDynamicControl(IDC_ERRORMESSAGE,			vgui::ControlResize_Stretch,	vgui::ControlResize_Move);

	AddDynamicControl(IDC_LABEL_PROGRESS,		vgui::ControlResize_Stretch,	vgui::ControlResize_Move);
	AddDynamicControl(IDC_PROGRESS,				vgui::ControlResize_Move,		vgui::ControlResize_Move);

	AddDynamicControl(IDC_GROUP_RENAME,			vgui::ControlResize_Stretch,	vgui::ControlResize_Move);
	AddDynamicControl(IDC_LABELPATTERN,			vgui::ControlResize_Static,		vgui::ControlResize_Move);
	AddDynamicControl(IDC_PATTERN,				vgui::ControlResize_Stretch,	vgui::ControlResize_Move);
	AddDynamicControl(IDC_LABELNEWNAME,			vgui::ControlResize_Static,		vgui::ControlResize_Move);
	AddDynamicControl(IDC_NEWNAME,				vgui::ControlResize_Stretch,	vgui::ControlResize_Move);
	AddDynamicControl(IDC_USEMETADATA,			vgui::ControlResize_Static,		vgui::ControlResize_Move);
	
	AddDynamicControl(IDC_OPTIONS,				vgui::ControlResize_Static,		vgui::ControlResize_Move);
	AddDynamicControl(IDC_UNDO,					vgui::ControlResize_MoveCentre,	vgui::ControlResize_Move);
	AddDynamicControl(IDC_CLEARERRORS,			vgui::ControlResize_MoveCentre,	vgui::ControlResize_Move);
	AddDynamicControl(IDC_RENAME,				vgui::ControlResize_MoveCentre,	vgui::ControlResize_Move);
	AddDynamicControl(IDOK,						vgui::ControlResize_Move,		vgui::ControlResize_Move);
}

/**************************************************************************

>	void FilenameList::ResizeFilenameColumns()

	Author:		Tim Browse
	Date:		5/5/2002

	Purpose:	Resize the original/new filename columns in the list box
				so that they take up half the area available.

**************************************************************************/

void F2Window::ResizeFilenameColumns()
{
	const int kColumnGap = vgui::Metrics::GetListBoxColumnGap();

	int StatusWidth = m_FileList.GetColumnWidth(kColumn_Status) + kColumnGap;

	vgui::Rect ListArea;
	GetControlPosition(IDC_FILELIST, &ListArea);
	int WidthLeft = (ListArea.GetWidth() - StatusWidth);

	int ScrollBarWidth = vgui::Metrics::GetVerticalScrollBarWidth();

	// Bodge for bloody Win32 listview controls! Gah!
	WidthLeft -= 1;
	
	int ColumnWidth1 = (WidthLeft - ScrollBarWidth) / 2;

	int ColumnWidth2 = WidthLeft - ColumnWidth1;

	if (m_FileList.IsScrollBarVisible(vgui::sb_Vertical))
	{
		ColumnWidth2 -= ScrollBarWidth;
	}

	m_FileList.SetColumnWidth(kColumn_OriginalName, ColumnWidth1 - kColumnGap);

	// We subtract 2 * kColumnGap from ColumnWidth2, as it is the last column,
	// so it has an end gap too.
	m_FileList.SetColumnWidth(kColumn_NewName, ColumnWidth2 - (2* kColumnGap));
}

/**************************************************************************

>	void F2Window::FillFileList()

	Author:		Tim Browse
	Date:		28/4/2002

	Purpose:	Fill in the list control with the names of all mp3 files
				found in the specified folder.

**************************************************************************/

void F2Window::FillFileList()
{
	// Empty list box
	m_FileList.Clear();

	m_FileList.LockRedraw();

	// Ok, populate that list box!
	uint cFiles = m_FilenameList.GetFilenameCount();

	vos::FileSpec fsFilename;

	vos::String EmptyString;

	for (int iFile = 0; iFile < cFiles; iFile++)
	{
		if (m_FilenameList.GetOriginalFilename(iFile, &fsFilename))
		{
			int iItem;

			// Add empty item...
			if (m_FileList.AddItem(EmptyString, &iItem))
			{
				// then fill it in
				m_FileList.SetItemText(iItem, "OK", kColumn_Status);
				m_FileList.SetItemText(iItem, fsFilename.GetLeafname(), kColumn_OriginalName);

				if (m_FilenameList.GetNewFilename(iFile, &fsFilename))
				{
					m_FileList.SetItemText(iItem, fsFilename.GetString(), kColumn_NewName);
				}

				// Set user data, to link back to FilenameList
				m_FileList.SetItemData(iItem, (vgui::UserData) iFile);
			}
		}
	}

	// Fill in the status column
	RefreshFilenameStatus();

	// Auto-fit columns, if user wants us to
	ResizeColumns();

	// Make sure list is sorted
	ForceFileListSort();

	// Finished changing the list control
	m_FileList.UnlockRedraw();
}

void F2Window::SetFolderOptions()
{
	// Tell filename list object what to scan
	m_FilenameList.SetFolder((vos::String) m_FolderName);
}

void F2Window::SetFilterOptions()
{
	// And how to filter it...
	if (m_UseFilter)
	{
		// Wildcard or RegExp?
		if (m_UseNormalFilter)
		{
			m_FilenameList.SetWildcardFilter(m_FileFilter, m_IgnoreCase);
		}
		else
		{
			m_FilenameList.SetRegExpFilter(m_FileFilter, m_IgnoreCase);
		}
	}
	else
	{
		// No filtering
		m_FilenameList.DisableFiltering();
	}
}

void F2Window::SetRenameOptions()
{
	// Tell filename list object how to rename
	vos::String Pattern = m_Pattern;
	if (Pattern.IsEmpty())
	{
		m_FilenameList.DisablePatternMatching();
	}
	else
	{
		m_FilenameList.SetPatternMatching(Pattern);
	}

	m_FilenameList.SetRenameSpec(m_NewName);
}


bool F2Window::MakeNewName(int iFileItem, vos::String* pNewName)
{
	// Get the name of the file
	vos::String Filename;
	if (!m_FileList.GetItemText(iFileItem, &Filename))
	{
		return false;
	}

	// Set up filename transform
	FilenameTransformSpec Spec;
	Spec.SetSourcePattern(m_Pattern);
	Spec.m_fsFolder = vos::FileSpec(m_FolderName);
	Spec.m_NewNameSpec = m_NewName;

	FilenameTransformer Transformer(Spec);
	Transformer.TransformFilename(Filename, pNewName);
	return true;
}

bool F2Window::OnFileSelectionChange()
{
	return true;
}


bool F2Window::OnPreview()
{
	ForcePreview(0);
	return true;
}

bool F2Window::OnRename()
{
	// if we're already doing bg processing, interrupt it
	CancelBackgroundProcessing();

	// Set up our options
	SetFolderOptions();
	SetFilterOptions();
	SetRenameOptions();

	// Get ready to rename files...
	vgui::ProgressDialog PreparationDialog(&m_FilenameList, IDS_PREPARETORENAMEFILES, this);

	if (!PreparationDialog.ShowProgressWithPercentage())
	{
		// User did not let us finish preparing
		vgui::Alert::Show(IDS_RENAMECANCELLED_TITLE, IDS_RENAMECANCELLED);
		return true;
	}

	// If the preparation process finished, did it work ok?
	if (!m_FilenameList.IsListValidated())
	{
		vgui::Alert::Show(IDS_RENAMEERROR_TITLE, IDS_RENAMEERROR);
		return true;
	}

	// Clear out existing undo history
	m_UndoManager.ClearHistory();

	// Make a worker object to rename the files
	FileRenameWorker RenameWorker(&m_FilenameList, &m_UndoManager);

	// Rename files, and display progress to user.
	vgui::ProgressDialog RenameDialog(&RenameWorker, IDS_RENAMINGFILES, this);
	if (RenameDialog.ShowProgressWithPercentage())
	{
		// Job not cancelled - but was it a success?
		if (m_FilenameList.GetErrorCount() == 0)
		{
			vgui::Alert::Show(IDS_RENAMECOMPLETE_TITLE, IDS_RENAMECOMPLETE, 
							  this, "UI.TellUserAbout.RenameSuccess");
		}
		else
		{
			// Errors occurred - tell user...
			vgui::Alert::Show(IDS_RENAMEERRORS_TITLE, IDS_RENAMEERRORS, 
							  this, "UI.TellUserAbout.RenameErrors");

			// Go bing, to let user know there were errors, even if they have
			// elected not to see the above alert dialog
			// TODO: go bing!

			// Disable rename for now, until they clear the errors.
			EnterErrorProgramState();
			RefreshFilenameStatus();
		}
	}
	else
	{
		// user cancelled the job.
	}

	// Ok, enable/disable Undo button according to what happened
	if (m_UndoManager.PossibleToUndo())
		EnableControl(IDC_UNDO);
	else
		DisableControl(IDC_UNDO);

	// Update any errors/success display
	RefreshFilenameStatus();

	return true;
}


void F2Window::RefreshFilenameStatus()
{
	// Does the user want to know about read-only files?
	bool bWarnAboutReadOnlyFiles = true;

	vos::String ReadOnlyPrefValue("Ask");
	vos::GetPreferences().Read("UI.AskUserAbout.ReadOnlyFiles", &ReadOnlyPrefValue);

	if (ReadOnlyPrefValue.CompareText("Always"))
	{
		// User has told us to always rename read-only files, so we
		// don't warn them about this
		bWarnAboutReadOnlyFiles = false;
	}

	vgui::UserData UserData;
	TRenameState Result;
	int iResult;

	int cFilenames = m_FileList.GetCount();

	for (int iItem = 0; iItem < cFilenames; iItem++)
	{
		// Get filename index for this item
		if (m_FileList.GetItemUserData(iItem, &UserData))
		{
			iResult = (int) UserData;

			if (m_FilenameList.GetRenameState(iResult, &Result))
			{
				// Special case for read-only warning - we don't show it if
				// user has told us they don't care about read-only files.
				if ((Result != RenameState_WarnReadOnly) ||
					bWarnAboutReadOnlyFiles)
				{
					// Just show status
					m_FileList.SetItemText(iItem, kResultText[Result].Load(), kColumn_Status);
				}
				else
				{
					// Hide warning - tell them it's OK
					m_FileList.SetItemText(iItem, kResultText[RenameState_OK].Load(), kColumn_Status);
				}
			}
			else
			{
				m_FileList.SetItemText(iResult, "<internal error>", kColumn_Status);
			}
		}
	}

	ResizeColumns();
	
	// Make sure list is sorted
	ForceFileListSort();
}

void F2Window::ResizeColumns()
{
	// Should we rejig the columns?
	if (m_bAutoSizeColumns)
	{
		m_FileList.AutoSizeColumn(kColumn_Status);
		ResizeFilenameColumns();
	}
}

void F2Window::EnterErrorProgramState()
{
	m_State = F2State_ErrorsOutstanding;
	DisableControl(IDC_RENAME);
	EnableControl(IDC_CLEARERRORS);
	ShowControl(IDC_ERRORMESSAGE);
}

void F2Window::RestoreNormalState()
{
	m_FilenameList.ClearErrors();
	m_State = F2State_RenameAllowed;
	EnableControl(IDC_RENAME);
	DisableControl(IDC_CLEARERRORS);
	HideControl(IDC_ERRORMESSAGE);
	FillFileList();
	ResizeColumns();

	// Get to work validating the names
	ForceValidate();	
}

bool F2Window::OnClearErrors()
{
	RestoreNormalState();
	return true;
}

bool F2Window::OnFilterFilesClicked()
{
	// Get state of the control
	if (m_UseFilter)
	{
		EnableControlGroup(kFilterGroup);
	}
	else
	{
		DisableControlGroup(kFilterGroup);
	}

	OnFilterOptionChanged();

	return true;
}

bool F2Window::OnRefresh()
{
	ForceFolderScan();
	return true;
}


bool F2Window::OnFilterOptionChanged()
{
	ForceFilter();
//	::MessageBeep(MB_OK);
	return true;
}

bool F2Window::OnBrowse()
{
	vgui::BrowseDialog Dialog(vgui::bdf_ObjectMustExist, this);

	vos::FileSpec NewFolder;
	if (Dialog.BrowseForFolder("Choose folder containing files to be renamed...", 
							   "", &NewFolder))
	{
		// Found a new folder...
		m_FolderName = NewFolder.GetString();
		ForceFolderScan();
	}

	return true;
}

bool F2Window::OnKeyDown(vgui::TKeyCode Key, int Modifiers)
{
	// Map function keys to buttons
	switch (Key)
	{
		case vgui::kc_F2:
			return OnRename();

		case vgui::kc_F5:
			return OnRefresh();
	}

	// Key not known
	return false;
}


bool F2Window::OnRenameSchemeChanged()
{
	ForcePreview();
	return true;
}

bool F2Window::OnFolderNameChanged()
{
	OnRefresh();
	return true;
}

bool F2Window::OnIdle(uint iIdleEvent)
{
	int PercentComplete = 0;
	if (m_FilenameList.DoWork(this, &PercentComplete) == vgui::Job_Finished)
	{
		// Hide progress display, and fill in the list
		HideProgressDisplay();
		FillFileList();

		// All done - we don't need more events
		ReleaseIdleEvents();

		m_bDoingBackgroundProcessing = false;
		return false;
	}
	else
	{
		// Still got work to do
		UpdateJobProgressDisplay(PercentComplete);
		return true;
	}
}

void F2Window::SetJobDescription(const vos::String& Description)
{
	m_JobDescription = Description;

	// If we have the full display on show (and not just "Working...")
	// then we update the text control.
	if (m_bFullProgressDisplayVisible)
		m_ProgressLabel.SetText(m_JobDescription);
}

void F2Window::UpdateJobProgressDisplay(uint PercentComplete)
{
	ShowProgressDisplay();
	m_ProgressBar.SetPosition(PercentComplete);
}

void F2Window::ShowProgressDisplay()
{
	// If the progress display is being revealed, then restart the
	// clock.
	if (!m_bProgressDisplayVisible)
	{
		ShowControl(IDC_LABEL_PROGRESS);

		// Resize the label control to extend as far right as the (hidden)
		// progress bar
		vgui::Rect LabelRect, ProgressBarRect;
		GetControlPosition(IDC_LABEL_PROGRESS, &LabelRect);
		GetControlPosition(IDC_PROGRESS, &ProgressBarRect);
		LabelRect.right = ProgressBarRect.right;

		ResizeControl(IDC_LABEL_PROGRESS, 
					  LabelRect.GetWidth(), LabelRect.GetHeight());

		// Ok, set 'working' message, and reset timestamp
		m_ProgressLabel.SetText("Working...");
		m_bProgressDisplayVisible = true;
		m_bFullProgressDisplayVisible = false;
		m_TimeJobStarted = vos::GetMonotonicTime();
	}

	if (!m_bFullProgressDisplayVisible)
	{
		// Ok, it's not visible, but have we passed the threshold yet?
		uint32 TimeNow = vos::GetMonotonicTime();

		if (TimeNow > (m_TimeJobStarted + kProgressTimeThreshold))
		{
			// Time to turn on full description/progress bar

			// Resize the label control to allow progress bar to be seen.
			vgui::Rect LabelRect, ProgressBarRect;
			GetControlPosition(IDC_LABEL_PROGRESS, &LabelRect);
			GetControlPosition(IDC_PROGRESS, &ProgressBarRect);

			LabelRect.right = ProgressBarRect.right - m_ProgressWidth;

			ResizeControl(IDC_LABEL_PROGRESS, 
						  LabelRect.GetWidth(), LabelRect.GetHeight());

			// Show progress bar and full job description
			ShowControl(IDC_PROGRESS);
			m_ProgressLabel.SetText(m_JobDescription);
			UpdateWindow();
			m_bFullProgressDisplayVisible = true;
		}
	}
}

void F2Window::HideProgressDisplay()
{
	if (m_bProgressDisplayVisible)
	{
		HideControlGroup(kProgressGroup);
		UpdateWindow();
		m_bProgressDisplayVisible = false;
		m_bFullProgressDisplayVisible = false;
	}
}

bool F2Window::IsJobCancelled() const
{
	return false;
}

bool F2Window::OnFilterCommit()
{
	// Refilter right now, with no delay
	ForceFilter(0);
	return true;
}


bool F2Window::OnRenameCommit()
{
	ForcePreview(0);
	return true;
}

void F2Window::ForceFolderScan()
{
	m_FilenameList.PrepareToStartProcessing(FilenameList::Work_Rescan);

	// For folder scans, we start right now...
	StartBackgroundProcess();
}

void F2Window::ForceFilter(uint msDelay)
{
	m_FilenameList.PrepareToStartProcessing(FilenameList::Work_Refilter);
	ScheduleBackgroundProcess(msDelay);
}

void F2Window::ForcePreview(uint msDelay)
{
	m_FilenameList.PrepareToStartProcessing(FilenameList::Work_Preview);
	ScheduleBackgroundProcess(msDelay);
}

void F2Window::ForceValidate(uint msDelay)
{
	m_FilenameList.PrepareToStartProcessing(FilenameList::Work_Validate);
	ScheduleBackgroundProcess(msDelay);
}

void F2Window::KillTimer()
{
	if (m_TimerID != vgui::NullTimer)
	{
		vgui::Timer::KillTimer(m_TimerID);
		m_TimerID = vgui::NullTimer;
	}
}

void F2Window::ScheduleBackgroundProcess(int msDelay)
{
	// Cancel any outstanding timer event
	KillTimer();

	// Ok, schedule a new timer for this window
	m_TimerID = vgui::Timer::SetTimer(vgui::TimerType_OneShot, msDelay, this);
}

bool F2Window::OnTimer(vgui::TimerID ID)
{
	if (ID == m_TimerID)
	{
		// This is a one-shot timer, so we won't get it again.
		//
		// NB. we reset timer ID here because otherwise StartBackgroundProcess()
		//	   will try to kill the timer, but it's already deregistered by vGUI.
		//	   (because it's a one-shot timer)
		//
		m_TimerID = vgui::NullTimer;

		// Ok, get processing
		StartBackgroundProcess();
	}
	else
	{
		COMPLAIN("Unknown timer event!");
	}

	return true;
}

void F2Window::StartBackgroundProcess()
{
	// Cancel any outstanding timer event
	KillTimer();

	// if we're already doing bg processing, interrupt it
	CancelBackgroundProcessing();

	// Set up our options
	SetFolderOptions();
	SetFilterOptions();
	SetRenameOptions();

	// Make sure we get a progress display
	m_FilenameList.EnableProgressDisplay(this);

	// Off we go!
	m_bDoingBackgroundProcessing = true;
	m_TimeJobStarted = vos::GetMonotonicTime();
	ClaimIdleEvents();
}

void F2Window::CancelBackgroundProcessing()
{
	if (m_bDoingBackgroundProcessing)
	{
		m_FilenameList.StopProcessing();
		HideProgressDisplay();
		m_bDoingBackgroundProcessing = false;
	}
}


bool F2Window::OnUndo()
{
	// Sanity check
	if (!m_UndoManager.PossibleToUndo())
	{
		vgui::Alert::Show("Undo...",
						  "There is currently nothing to undo.");
		DisableControl(IDC_UNDO);
		return true;
	}

	// Ask them if they're sure!
	const vgui::StringResourceID rButtons[] = 
	{
		IDS_BUTTON_YES, IDS_BUTTON_NO, NULL
	};

	vgui::Question UndoQuestion(IDS_UNDO_TITLE, IDS_UNDO_WARNING);
	UndoQuestion.SetCheckBoxPreference("UI.AskUserAbout.Undo");
	
	if (UndoQuestion.Ask(rButtons) != IDS_BUTTON_YES)
	{
		// User does not want to undo
		return true;
	}

	// Ok, start the undo procedure...
	m_UndoManager.PrepareToUndo();
	vgui::ProgressDialog Dialog(&m_UndoManager, "Undoing file rename operation...", this);
	if (Dialog.ShowProgressWithoutCancel())
	{
		vgui::Alert::Show(IDS_UNDOCOMPLETE_TITLE, IDS_UNDOCOMPLETE,
						  this, "UI.TellUserAbout.UndoSuccess");
	}
	else
	{
		vgui::Alert::Show(IDS_UNDOERRORS_TITLE, IDS_UNDOERRORS);
	}

	// And we don't want to try that again
	m_UndoManager.ClearHistory();
	DisableControl(IDC_UNDO);

	// Make sure list results are up to date
	RefreshFilenameStatus();

	// Revalidate filenames in background (so that read-only files
	// can be detected and shown again)
	ForceValidate();

	return true;
}

bool F2Window::OnResize(int NewWidth, int NewHeight)
{
	// Let base Dialog class do its clever stuff first...
	Dialog::OnResize(NewWidth, NewHeight);

	// Resize file list columns, if required
	if (m_bAutoSizeColumns)
		ResizeFilenameColumns();

	return true;
}

int F2Window::ListSortFunction(const vgui::ListControl* pListControl, 
							   vgui::UserData TheUserData,
							   int iItem1, int iItem2, 
							   int iColumn)
{
	// Get access to the FilenameList.
	F2Window *pF2Window = (F2Window*) TheUserData;
	FilenameList* pFilenameList = &pF2Window->m_FilenameList;

	// Ok, extract the indices for the two items in question
	vgui::UserData UserData;

	pListControl->GetItemUserData(iItem1, &UserData);
	int iFilename1 = (int) UserData;

	pListControl->GetItemUserData(iItem2, &UserData);
	int iFilename2 = (int) UserData;

	// Check for enforced status sort first
	TRenameState State1, State2;

	int StateOrder = 0;

	if (pFilenameList->GetRenameState(iFilename1, &State1) &&
		pFilenameList->GetRenameState(iFilename2, &State2))
	{
		// We use this order to show errors first, then warnings, then success
		if (State1 < State2)
			StateOrder = -1;
		else if (State1 > State2)
			StateOrder = 1;

		// If sort order is inverted, then we also invert, as we always
		// want to sort with errors first, then warnings, etc
		if (pListControl->GetSortOrder() == vgui::ListControl::SortOrder_Descending)
		{
			StateOrder = -StateOrder;
		}
	}
	else
	{
		COMPLAIN("Unable to get rename state when sorting!");
		return 0;
	}

	if (pF2Window->m_bSortByStatusFirst && (StateOrder != 0))
	{
		return StateOrder;
	}

	// Ok, now sort on the specified column
	if (iColumn == kColumn_Status)
	{
		// Special case for state, as we've already computed it.
		return StateOrder;
	}

	// Ok, we're sorting on filenames, so get the relevant filenames
	vos::FileSpec FS1, FS2;

	if (iColumn == kColumn_OriginalName)
	{
		if (!pFilenameList->GetOriginalFilename(iFilename1, &FS1) ||
			!pFilenameList->GetOriginalFilename(iFilename2, &FS2))
		{
			// Failed to get string - assume equality
			return 0;
		}
	}
	else if (iColumn == kColumn_NewName)
	{
		if (!pFilenameList->GetNewFilename(iFilename1, &FS1) ||
			!pFilenameList->GetNewFilename(iFilename2, &FS2))
		{
			// Failed to get string - assume equality
			return 0;
		}
	}
	else
	{
		COMPLAIN("Unknown column sort request in ListSortFunction()!");
		return 0;
	}

	// Convert FileSpec to string, and sort
	vos::String S1(FS1.GetString());
	vos::String S2(FS2.GetString());
	return S1.CompareOrder(S2);
}

bool F2Window::OnSortChange()
{
	// Cache value of checkbox (so we don't have to read it on every
	// sort operation).
	m_bSortByStatusFirst = m_SortOnStatus;
	ForceFileListSort();
	return true;
}

void F2Window::ForceFileListSort()
{
	m_FileList.EnableSorting();
}

bool F2Window::OnSortNow()
{
	ForceFileListSort();
	return true;
}

bool F2Window::OnOptions()
{
	OptionsDialog Dialog(this);
	Dialog.DoModal();
	return true;
}
