//
// FileRenameWorker.cpp
//

//
// This software is protected by the GNU General Public License,
// also known as the GPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "FileRenameWorker.h"

#include "FilenameList.h"

#include "vgui\Question.h"

#include "vos\File.h"
#include "vos\Preferences.h"

#include "Res\Resource.h"

FileRenameWorker::FileRenameWorker(FilenameList* pFilenameList,
								   IFileRenamerMonitor* pMonitor)
								   
	  : m_pFilenameList(pFilenameList),
		m_cFiles(0),
		m_iCurrentFile(0),
		m_ReadOnlyAction(ReadOnlyAction_Ask)

{
	// Find out how many files we are to process.
	m_cFiles = m_pFilenameList->GetFilenameCount();
	m_iCurrentFile = 0;

	// Install rename monitor
	m_Renamer.SetMonitor(pMonitor);

	// Tell renamer where files are
	m_Renamer.SetFolder(m_pFilenameList->GetFolder());

	// Clear any existing errors from filename list
	m_pFilenameList->ClearErrors();
	
	// Read the preference for dealing with read-only files
	vos::String ReadOnlyPref("Ask");
	vos::GetPreferences().Read("UI.AskUserAbout.ReadOnlyFiles", &ReadOnlyPref);

	if (ReadOnlyPref.CompareText("Always"))
	{
		// We always rename read-only files
		m_ReadOnlyAction = ReadOnlyAction_Rename;
	}
	else if (ReadOnlyPref.CompareText("Never"))
	{
		// We never rename read-only files
		m_ReadOnlyAction = ReadOnlyAction_Ignore;
	}
	else
	{
		// Otherwise, we always ask the user what to do
		m_ReadOnlyAction = ReadOnlyAction_Ask;
	}
}

vgui::TJobStatus FileRenameWorker::DoWork(vgui::JobController* pJobController,
										  int* pPercentComplete)
{
	int cFilesToProcess = 25;
	vos::FileSpec fsOriginalFilename, 
				  fsNewFilename;

	TRenameState State;

	bool bRenameFile;

	while ((cFilesToProcess > 0) && (m_iCurrentFile < m_cFiles))
	{
		// Should we rename this file?
		bRenameFile = false;

		vos::FileSpec fsOriginalFilename;
		m_pFilenameList->GetOriginalFilename(m_iCurrentFile, &fsOriginalFilename);

		// Check to see if file is read-only
		if (fsOriginalFilename.IsReadOnly())
		{
			// Ok, update the rename status on this file
			m_pFilenameList->MarkFilenameProblem(m_iCurrentFile, RenameState_WarnReadOnly);
		}

		if (m_pFilenameList->GetRenameState(m_iCurrentFile, &State))
		{
			// Accept any files without warnings
			if (State == RenameState_OK)
			{
				bRenameFile = true;
			}
			else if (State == RenameState_WarnReadOnly)
			{
				// File is read-only - find out what we should do about it.
				if (m_ReadOnlyAction == ReadOnlyAction_Rename)
				{
					// Silently rename all read-only files
					bRenameFile = true;
				}
				else if (m_ReadOnlyAction == ReadOnlyAction_Ask)
				{
					// Ask the user what they want to do about this file...
					if (AskUserAboutReadOnlyFile(fsOriginalFilename.GetString()))
					{
						// user wants to rename the file
						bRenameFile = true;
					}
				}
				else
				{
					// Otherwise, ignore read-only files
				}
			}
		}

		// Ok, so should we rename it?
		if (bRenameFile)
		{
			// Yes, so get the filenames
			if (m_pFilenameList->GetOriginalFilename(m_iCurrentFile, &fsOriginalFilename) &&
				m_pFilenameList->GetNewFilename(m_iCurrentFile, &fsNewFilename))
			{
				TRenameState State = RenameState_RenamedOK;
				
				if (!m_Renamer.RenameFile(m_iCurrentFile, 
										  fsOriginalFilename, 
										  fsNewFilename))
				{
					// Error occurred...
					switch (vos::GetLastErrorCode())
					{
						case vos::Err_File_NotFound:
							State = RenameState_ErrorFileNotFound;
							break;

						case vos::Err_File_AccessDenied:
							State = RenameState_ErrorAccessDenied;
							break;

						case vos::Err_File_AlreadyExists:
							State = RenameState_ErrorFileAlreadyExists;
							break;

						case vos::Err_File_CreateFailed:
							State = RenameState_ErrorFolderCreateFailed;
							break;

						default:
							State = RenameState_ErrorUnknown;
							break;
					}
						
					// Error handled - clear it
					ERROR_HANDLED();
				}

				// Update new state of filename
				if (State == RenameState_RenamedOK)
				{
					m_pFilenameList->SetFilenameStatus(m_iCurrentFile, State);
				}
				else
				{
					m_pFilenameList->MarkFilenameProblem(m_iCurrentFile, State);
				}
			}
			else
			{
				// Tell user something went wrong, and we're very sorry
				State = RenameState_ErrorInternal;
			}
		}
		else
		{
			// Leave filename state showing warning
		}

		// Try next file
		m_iCurrentFile++;
		cFilesToProcess--;
	}

	// Tell progress dialog how we're doing.
	if (m_cFiles > 0)
		*pPercentComplete = (100 * m_iCurrentFile) / m_cFiles;
	else
		*pPercentComplete = 100;

	// And tell them if we've finished yet
	if (m_iCurrentFile < m_cFiles)
	{
		// Still more files to rename
		return vgui::Job_NotFinished;
	}
	else
	{
		// All done
		return vgui::Job_Finished;
	}
}

void FileRenameWorker::OnCancel(vgui::JobController* pJobController)
{
	// TODO: Ask the user if they want to undo files processed so far...
}

bool FileRenameWorker::AskUserAboutReadOnlyFile(const vos::String& Filename)
{
	// Ask the user what they want to do about this file...
	vgui::StringResourceID rMessageFormat(IDS_RENAMEREADONLY);
	vgui::StringResourceID rMessageTitle(IDS_RENAMEREADONLY_TITLE);

	vos::String MessageTitle = rMessageTitle.Load();

	// Format the message string, inserting filename
	vos::String MessageFormat(rMessageFormat.Load());
	vos::String Message;
	Message.Printf(MessageFormat, Filename.c_str());

	// Array of buttons for the question dialog
	static vgui::StringResourceID rButtons[] = 
	{
		IDS_BUTTON_YES,
		IDS_BUTTON_NO,
		IDS_BUTTON_YESALL,
		IDS_BUTTON_NOALL,
		0
	};

	// Ok, now ask the user
	vgui::Question ReadOnlyQuestion(MessageTitle.c_str(), Message.c_str());
	ReadOnlyQuestion.EnableCheckBox();
	vgui::StringResourceID Response = ReadOnlyQuestion.Ask(rButtons);

	const char* szPrefName = "UI.AskUserAbout.ReadOnlyFiles";

	if (Response == IDS_BUTTON_YES)
	{
		// User wants to rename file - do they want to be asked again?
		if (!ReadOnlyQuestion.AskUserNextTime())
		{
			// No, so change action/preference accordingly
			m_ReadOnlyAction = ReadOnlyAction_Rename;
			vos::GetPreferences().Write(szPrefName, "Always");
		}
		return true;
	}
	else if (Response == IDS_BUTTON_NO)
	{
		// User does not want to rename filee - do they want to be asked again?
		if (!ReadOnlyQuestion.AskUserNextTime())
		{
			// No, so change action/preference accordingly
			m_ReadOnlyAction = ReadOnlyAction_Ignore;
			vos::GetPreferences().Write(szPrefName, "Never");
		}
		return false;
	}
	else if (Response == IDS_BUTTON_YESALL)
	{
		// user wants to rename this file, and all read-only files
		m_ReadOnlyAction = ReadOnlyAction_Rename;

		// If user does not want to be asked again, change up the preference
		if (!ReadOnlyQuestion.AskUserNextTime())
		{
			vos::GetPreferences().Write(szPrefName, "Always");
		}

		return true;
	}
	else if (Response == IDS_BUTTON_NOALL)
	{
		// user does not want to rename this file, or any other read-only
		// files.
		m_ReadOnlyAction = ReadOnlyAction_Ignore;

		// If user does not want to be asked again, change up the preference
		if (!ReadOnlyQuestion.AskUserNextTime())
		{
			vos::GetPreferences().Write(szPrefName, "Never");
		}

		return false;
	}
	else
	{
		COMPLAIN("Logic error - unknown question response!");

		// Play it safe, and don't rename this file or any others
		m_ReadOnlyAction = ReadOnlyAction_Ignore;
		return false;
	}
}
