//
// FileRenamer.cpp
//

//
// This software is protected by the GNU General Public License,
// also known as the GPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "FileRenamer.h"

#include "vgui\Errors.h"

#include "vos\File.h"
#include "vos\Trace.h"

#include "Res\Resource.h"

#include "IFileRenamerMonitor.h"

/**************************************************************************

>	FileRenamer::FileRenamer()

	Author:		Tim Browse
	Date:		29/4/2002

	Purpose:	Constructor

**************************************************************************/

FileRenamer::FileRenamer() : m_pMonitor(NULL)
{
	// TCILB
}

/**************************************************************************

>	void FileRenamer::SetFolder(const vos::FileSpec& fsFolder)

	Author:		Tim Browse
	Date:		3/5/2002
	Inputs:		fsFolder - location of the files to be renamed.

	Purpose:	Tell the renamer object where the files are located.

**************************************************************************/

void FileRenamer::SetFolder(const vos::FileSpec& fsFolder)
{
	m_fsFolder = fsFolder;
}

/**************************************************************************

>	void FileRenamer::SetMonitor(IFileRenamerMonitor* pMonitor)

	Author:		Tim Browse
	Date:		3/5/2002
	Inputs:		pMonitor - the monitor object to use

	Purpose:	Tell the renamer what monitor object to use.  This object
				will be notified of any file renames/folder changes that
				are made during a file rename.

				See IFileRenamerMonitor interface for details.

**************************************************************************/

void FileRenamer::SetMonitor(IFileRenamerMonitor* pMonitor)
{
	m_pMonitor = pMonitor;
}

/**************************************************************************

>	vos::Error FileRenamer::RenameFile(uint iFileIndex,
									   const vos::FileSpec& fsFileToRename,
									   const vos::FileSpec& fsNewFilename)

	Author:		Tim Browse
	Date:		29/4/2002
	Inputs:		iFileIndex - index/ID of the file being renamed (used to
							 refer back to the FilenameList object)
				fsFileToRename - the file that should be renamed.
				fsNewFilename - the new name to give the file
	Returns:	Any errors that might occur during the process.

	Purpose:	Rename the named file, according the the transform spec
				this object is using.

				If the file changes directory, any required directories
				will be created before the file is renamed to the new
				location.

**************************************************************************/

vos::Error FileRenamer::RenameFile(uint iFileIndex,
								   const vos::FileSpec& fsFileToRename,
								   const vos::FileSpec& fsNewFilename)
{
	// Wipe our list of new folders
	m_NewFolders.clear();

	// Ok, find out if we need to create any new folders.
	// NB. We use main folder to create full pathname when checking
	//	   for parent folders (otherwise it won't work!)
	vos::FileSpec fsFullNewFilename(m_fsFolder);
	fsFullNewFilename.AppendRelativeFileSpec(fsNewFilename);

	vos::FileSpec fsParentFolder;
	if (fsFullNewFilename.GetParent(&fsParentFolder))
	{
		// Does this folder exist?
		switch (fsParentFolder.GetFileType())
		{
			case vos::FileType_File:
			{
				// Destination folder is actually a file - we can't do this
				return ERROR_RES(vos::Err_File_AlreadyExists, IDS_FOLDEREXISTSASFILE);
			}

			case vos::FileType_NotFound:
			{
				// Destination folder does not exist, so try to create it
				// (and we keep track of any folders that are created).
				if (!fsParentFolder.CreateFolderTree(this))
				{
					// Could not create parent folder
					return LASTERROR;
				}

				break;
			}

			case vos::FileType_Folder:
			{
				// Destination folder exists - we do nothing
				break;
			}
		}
	}

	// Ok, parent folder exists, so we rename the file now...
	vos::FileSpec fsFullFilenameToRename(m_fsFolder);
	fsFullFilenameToRename.AppendRelativeFileSpec(fsFileToRename);

	if (fsFullFilenameToRename.Rename(fsFullNewFilename))
	{
		// Rename completed successfully
		if (m_pMonitor != NULL)
		{
			m_pMonitor->NotifyFileRenamed(iFileIndex, 
										  fsFullFilenameToRename, 
										  fsFullNewFilename);
		}

		return VNOERROR;
	}
	else
	{
		// Rename failed - remove any folders we just created to 
		// allow this file to be renamed
		CleanUpFolders();

		// Let caller know what went wrong.
		return LASTERROR;
	}
}

/**************************************************************************

>	void FileRenamer::CleanUpFolders()

	Author:		Tim Browse
	Date:		29/4/2002

	Purpose:	Remove any folders that were just created to contain
				the new filename.

**************************************************************************/

void FileRenamer::CleanUpFolders()
{
	if (!m_NewFolders.empty())
	{
		// Save error state so we don't corrupt the original error
		// (from the rename operation).
		ERRMGR().PushState();
		{
			// We remove folders in reverse order, because if we create:
			//
			//   1. folder
			//   2. folder\subfolder
			//
			// then we need to remove subfolder before removing folder...
			//
			std::vector<vos::FileSpec>::iterator it = m_NewFolders.end();

			while (it != m_NewFolders.begin())
			{
				// Ok, remove this folder...
				if (it->Delete())
				{
					// Folder deleted ok
					if (m_pMonitor != NULL)
					{
						m_pMonitor->NotifyFolderDeleted(*it);
					}
				}
				else
				{
					// For some reason we can't delete this folder - ignore
					// this (any monitor installed should notice the
					// discrepancy and report it to the user).
					ERROR_HANDLED();
				}

				--it;
			}
		}
		ERRMGR().PopState();
	}

	m_NewFolders.clear();
}


/**************************************************************************

>	void FileRenamer::NotifyFolderCreated(const vos::FileSpec& NewFolderName)

	Author:		Tim Browse
	Date:		29/4/2002
	Inputs:		NewFolderName - the name of the new folder that has just
								been created.

	Purpose:	Let the FileSpec class tell us when it creates a folder.

**************************************************************************/

void FileRenamer::NotifyFolderCreated(const vos::FileSpec& NewFolderName)
{
	// Add this to our list
	m_NewFolders.push_back(NewFolderName);

	// And tell our monitor, if one is installed
	if (m_pMonitor != NULL)
	{
		m_pMonitor->NotifyFolderCreated(NewFolderName);
	}
}
