//
// FilenameList.cpp
//

//
// This software is protected by the GNU General Public License,
// also known as the GPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "FilenameList.h"

#include "vos\FileMatches.h"
#include "vos\MonotonicTime.h"
#include "vos\Trace.h"

#include "vgui\ProgressDialog.h"

#include "ScanFilesWorker.h"
#include "FilterFilesWorker.h"
#include "GenerateFilenamesWorker.h"
#include "FilenameValidateWorker.h"


/**************************************************************************

>	FilenameList::FilenameList()

	Author:		Tim Browse
	Date:		30/4/2002

	Purpose:	Constructor

**************************************************************************/

FilenameList::FilenameList() : m_bUseFilter(false),
							   m_pParentWindow(NULL),
							   m_pWorker(NULL),
							   m_WorkStatus(Work_Rescan),
							   m_bIgnoreCaseInFilter(true),
							   m_cErrors(0),
							   m_cWarnings(0)
{
	// TCILB
}

/**************************************************************************

>	void FilenameList::SetFolder(const vos::FileSpec& fsFolderName)

	Author:		Tim Browse
	Date:		30/4/2002
	Inputs:		fsFolderName - the folder to scan for files.

	Purpose:	Tell the object which folder to scan for files.

				This will always cause the folder to be rescanned.

**************************************************************************/

void FilenameList::SetFolder(const vos::FileSpec& fsFolderName)
{
	// Ignore if no change
	if (fsFolderName != m_fsFolderName)
	{
		// Use new folder
		m_fsFolderName = fsFolderName;

		// We need to rescan/filter before we have valid data.
		PrepareToStartProcessing(Work_Rescan);
	}
}

/**************************************************************************

>	void FilenameList::DisableFiltering()

	Author:		Tim Browse
	Date:		30/4/2002

	Purpose:	Disable file filtering - all files in the folder being
				scanned will be added.

**************************************************************************/

void FilenameList::DisableFiltering()
{
	if (m_bUseFilter)
	{
		// Clear filter, and make sure we refilter (with no restrictions)
		m_bUseFilter = false;
		PrepareToStartProcessing(Work_Refilter);
	}
}

/**************************************************************************

>	void FilenameList::SetWildcardFilter(const vos::String& Filter,
										 bool bIgnoreCase = true)

	Author:		Tim Browse
	Date:		30/4/2002
	Inputs:		Filter - the new filter string to use, e.g. "*.mp3"
				bIgnoreCase - true => filter matching is case-insensitive

	Purpose:	Enable wildcard filtering - i.e. filters containing the
				usual filename wildcards ? and *.

**************************************************************************/

void FilenameList::SetWildcardFilter(const vos::String& Filter,
									 bool bIgnoreCase)
{
	// Turn filter into regexp
	vos::String RegExp = Filter;

	// Escape any special characters
	RegExp.Replace(".", "\\.");
	RegExp.Replace(")", "\\)");
	RegExp.Replace("(", "\\(");

	// Now convert file wildcards into regexp syntax
	RegExp.Replace("*", ".*");
	RegExp.Replace("?", ".");

	// Don't bother doing anything if there's no change
	if ((m_strFilter == RegExp) && (m_bIgnoreCaseInFilter == bIgnoreCase))
	{
		return;
	}

	// And use it
	try
	{
		m_Filter.SetExpression(RegExp.c_str(), bIgnoreCase);
	}
	catch(...)
	{
		// Bad regexp - ignore it for now
	}
	
	// Make sure we refilter
	m_bUseFilter = true;
	PrepareToStartProcessing(Work_Refilter);

	// Remember current filter for next time
	m_strFilter = RegExp;
	m_bIgnoreCaseInFilter = bIgnoreCase;
}

/**************************************************************************

>	void FilenameList::SetRegExpFilter(const vos::String& RegExp,
									   bool bIgnoreCase = true)

	Author:		Tim Browse
	Date:		30/4/2002
	Inputs:		RegExp - the regular expression filter to use
				bIgnoreCase - true => filter matching is case-insensitive

	Purpose:	Enable regular expression filtering - i.e. only files
				with names matching the specified regular expression
				will be added to the list.

**************************************************************************/

void FilenameList::SetRegExpFilter(const vos::String& RegExp,
								   bool bIgnoreCase)
{
	// Don't bother doing anything if there's no change
	if ((m_strFilter == RegExp) && (m_bIgnoreCaseInFilter == bIgnoreCase))
	{
		return;
	}

	// Use new filter...
	try 
	{
		m_Filter.SetExpression(RegExp.c_str(), bIgnoreCase);
	}
	catch(...)
	{
		// Bad regexp - ignore it for now
	}

	// ...and make sure we refilter
	m_bUseFilter = true;
	PrepareToStartProcessing(Work_Refilter);

	// Remember current filter for next time
	m_strFilter = RegExp;
	m_bIgnoreCaseInFilter = bIgnoreCase;
}

/**************************************************************************

>	uint FilenameList::GetFilenameCount()

	Author:		Tim Browse
	Date:		30/4/2002
	Returns:	The number of files in the filename list.

	Purpose:	Find out how many files are currently in the list.  This
				will always be an accurate representation, given the current
				folder name and any file filter that has been set up.

**************************************************************************/

uint FilenameList::GetFilenameCount()
{
	// Make sure we have filtered the filenames
	if (!IsListFiltered())
		return 0;

	// Tell them how many files we have, after filtering
	return m_FilteredFilenames.size();
}

/**************************************************************************

>	bool FilenameList::GetOriginalFilename(uint iFilename, vos::FileSpec* pfsFilename)

	Author:		Tim Browse
	Date:		30/4/2002
	Inputs:		iFilename - the filename to retrieve
	Outputs:	pfsFilename - the requested filename
	Returns:	true => filename was found
				false => index was out of bounds.

	Purpose:	Get the i'th filename in the list.

**************************************************************************/

bool FilenameList::GetOriginalFilename(uint iFilename, vos::FileSpec* pfsFilename)
{
	if (iFilename < m_FilteredFilenames.size())
	{
		// Go through filter array to find correct filename
		*pfsFilename = m_Filenames[m_FilteredFilenames[iFilename]];
		return true;
	}

	// Index out of bounds
	return false;
}

/**************************************************************************

>	bool FilenameList::GetNewFilename(uint iFilename, vos::FileSpec* pfsFilename)

	Author:		Tim Browse
	Date:		2/5/2002
	Inputs:		iFilename - the filename to retrieve
	Outputs:	pfsFilename - the requested filename
	Returns:	true => filename was found
				false => index was out of bounds.

	Purpose:	Get the new name for the i'th filename in the list.

**************************************************************************/

bool FilenameList::GetNewFilename(uint iFilename, vos::FileSpec* pfsFilename)
{
	if (iFilename < m_NewFilenames.size())
	{
		// Go through filter array to find correct filename
		*pfsFilename = m_NewFilenames[iFilename];
		return true;
	}

	// Index out of bounds
	COMPLAIN("Out of bounds index");
	return false;
}


/**************************************************************************

>	bool FilenameList::FileMatchesFilter(const vos::FileSpec& fsFilename)

	Author:		Tim Browse
	Date:		30/4/2002
	Inputs:		fsFilename - the filename to check against the filter.
	Returns:	true => filename matches filter
				false => otherwise

	Purpose:	Check to see if the given filename matches the current
				filter options.

**************************************************************************/

bool FilenameList::FileMatchesFilter(const vos::FileSpec& fsFilename)
{
	if (m_bUseFilter)
	{
		// Do regexp match on this filename
		return m_Filter.Match(fsFilename.GetString().c_str());
	}
	else
	{
		// No filtering - file will always match
		return true;
	}
}


void FilenameList::PrepareToStartProcessing(TWorkType WorkType)
{
	// We leave any current processing going - because the caller may
	// be asking for a refilter when we haven't finished scanning yet...
	if (WorkType < m_WorkStatus)
	{
		m_WorkStatus = WorkType;
	}
}

vgui::TJobStatus FilenameList::DoWork(vgui::JobController* pJobController,
									  int *pPercentComplete)
{
	// If we have no work to do at present, what should we be doing?
	if (m_pWorker == NULL)
	{
		switch (m_WorkStatus)
		{
			// Do we need to scan the files?
			case Work_Rescan:
			{
				// Yes, so scan them (with modal dialog), and return
				pJobController->SetJobDescription("Scanning files...please wait");
				pJobController->HideProgressDisplay();

				if (ScanFiles())
				{
					// Files scanned ok, so now we need to do filter/preview
					MoveToNextPhase();
					return vgui::Job_NotFinished;
				}
				else
				{
					// Failed to scan files - so we can't do anymore
					return vgui::Job_Finished;
				}
			}
			break;

			case Work_Refilter:
			{
				// Ok, time to start filtering files...
				m_pWorker = new FilterFilesWorker(this);
				pJobController->SetJobDescription("Filtering files...");
				pJobController->UpdateJobProgressDisplay(0);
			}
			break;

			case Work_Preview:
			{
				// Ok, time to start generating filenames...
				InitialiseResults();
				m_pWorker = new GenerateFilenamesWorker(this);
				pJobController->SetJobDescription("Generating filenames...");
				pJobController->UpdateJobProgressDisplay(0);
			}
			break;

			case Work_Validate:
			{
				// Ok, time to start validating filenames...
				m_pWorker = new FilenameValidateWorker(this);
				pJobController->SetJobDescription("Validating filenames...");
				pJobController->UpdateJobProgressDisplay(0);
			}
			break;
		}
	}

	if (m_pWorker != NULL)
	{
		// We have some work to do, so do it...
		vgui::TJobStatus Status = m_pWorker->DoWork(pJobController, pPercentComplete);

		// If this worker has finished, get rid of it
		if (Status == vgui::Job_Finished)
		{
			delete m_pWorker;
			m_pWorker = NULL;

			MoveToNextPhase();
		}

		// Assume there is more to do...
		return vgui::Job_NotFinished;
	}
	else
	{
		// No more work to do
		return vgui::Job_Finished;
	}
}

void FilenameList::StopProcessing()
{
	// Clean up any worker object
	if (m_pWorker != NULL)
	{
		delete m_pWorker;
		m_pWorker = NULL;
	}

	// Leave the m_WorkStatus variable as is, so we can pick up
	// from where necessary next time, and they reflect the validity
	// of the data for when IsDataValid() is called.
}

bool FilenameList::IsListFiltered()
{
	return (m_WorkStatus > Work_Refilter);
}

bool FilenameList::IsListValidated()
{
	return (m_WorkStatus == Work_Complete);
}

bool FilenameList::IsListReady()
{
	// Filename lists are ready => data is ready
	// (status may not be correct yet)
	return (m_WorkStatus >= Work_Validate);
}

/**************************************************************************

>	void FilenameList::EnableProgressDisplay(vgui::Window* pParent)

	Author:		Tim Browse
	Date:		30/4/2002
	Inputs:		pParent - the parent window for displaying child progress 
						  dialog.

	Purpose:	Enable progress display when scanning files.

**************************************************************************/

void FilenameList::EnableProgressDisplay(vgui::Window* pParent)
{
	m_pParentWindow = pParent;
}

/**************************************************************************

>	void FilenameList::DisableProgressDisplay()

	Author:		Tim Browse
	Date:		30/4/2002

	Purpose:	Disable progress display when scanning files.

**************************************************************************/

void FilenameList::DisableProgressDisplay()
{
	m_pParentWindow = NULL;
}

/**************************************************************************

>	bool FilenameList::ScanFiles()

	Author:		Tim Browse
	Date:		30/4/2002
	Returns:	true => folder was scanned ok
				false => otherwise (e.g. no folder set up yet)

	Purpose:	Scan the current folder, and read the names of all files
				in that folder.

**************************************************************************/

bool FilenameList::ScanFiles()
{
	// Clear the lists before we start
	m_Filenames.clear();
	m_FilteredFilenames.clear();
	m_NewFilenames.clear();

	// Check the folder exists first
	if (m_fsFolderName.GetFileType() != vos::FileType_Folder)
	{
		// Can't scan a file, or a non-existent folder
		return false;
	}

	// Set up a worker object to scan the files, so we can show a progress
	// display if required.
	ScanFilesWorker Worker(m_fsFolderName, &m_Filenames);

	if (m_pParentWindow != NULL)
	{
		// Show a progress display
		vgui::ProgressDialog Dialog(&Worker, "Scanning folder for files...", m_pParentWindow);
		Dialog.ShowProgress();
	}
	else
	{
		// No progress display - just do the job
		Worker.DoAllWork();
	}

	// Tell caller if we completed the scan ok
	return Worker.CompletedOK();	
}


void FilenameList::MoveToNextPhase()
{
	static TWorkType NextPhase[] = 
	{
		Work_Refilter,
		Work_Preview,
		Work_Validate,
		Work_Complete,
		Work_Complete
	};

	m_WorkStatus = NextPhase[m_WorkStatus];
}

vos::FileSpec FilenameList::GetFolder()
{
	return m_fsFolderName;
}

void FilenameList::DisablePatternMatching()
{
	m_bUsePatternMatching = false;
}

void FilenameList::SetPatternMatching(const vos::String& Pattern, 
									  bool bIgnoreCase)
{
	m_bUsePatternMatching = true;
	m_Pattern = Pattern;
}

void FilenameList::SetRenameSpec(const vos::String& RenameSpec)
{
	m_RenameSpec = RenameSpec;
}

void FilenameList::InitialiseResults()
{
	// Make sure we have enough space for results for the filenames we have
	m_Results.clear();
	uint cFilenames = GetFilenameCount();
	m_Results.resize(cFilenames);
	m_cErrors = 0;

	// Set all results to OK
	for (int iResult = 0; iResult < cFilenames; iResult++)
	{
		m_Results[iResult] = RenameState_OK;
	}
}

void FilenameList::ClearErrors()
{
	// We will count the warnings as we go...
	m_cWarnings = 0;

	int cResults = m_Results.size();

	for (int iResult = 0; iResult < cResults; iResult++)
	{
		// Find the actual filename associated with this result
		int iFilename = m_FilteredFilenames[iResult];

		switch (m_Results[iResult])
		{
			case RenameState_ErrorFileNotFound:
			{
				// File was not found - check to see if it now
				// exists...
				if (m_Filenames[iFilename].GetFileType() == vos::FileType_File)
				{
					// File exists, so we'll have a go again
					m_Results[iResult] = RenameState_OK;
				}
				else
				{
					// File not found - we won't try again
					m_Results[iResult] = RenameState_WarnFileNotFound;
				}
				break;
			}

			case RenameState_ErrorFileAlreadyExists:
			{
				// New file already exists - check to see if it now
				// exists...
				if (m_NewFilenames[iFilename].GetFileType() == vos::FileType_NotFound)
				{
					// New name does not exists, so we'll have a go again
					m_Results[iResult] = RenameState_OK;
				}
				else
				{
					// New filename already exists (as a file or folder),
					// so we don't try again.
					m_Results[iResult] = RenameState_WarnNewFileExists;
				}
				break;
			}

			// We'll deal with these as errors come up again (if they do)
			case RenameState_ErrorAccessDenied:
			case RenameState_ErrorFolderCreateFailed:
			case RenameState_ErrorUnknown:
			{
				m_Results[iResult] = RenameState_OK;
				break;
			}

			default:
			{
				// We leave any other results alone...
				break;
			}
		}

		// Accumulate warning total
		switch (m_Results[iResult])
		{
			case RenameState_WarnNewFileExists:
			case RenameState_WarnFileNotFound:
			case RenameState_WarnNameClash:
			{
				m_cWarnings++;
				break;
			}
		}
	}

	// No errors in the list now (but possibly some warnings)
	m_cErrors = 0;
}

void FilenameList::MarkFilenameProblem(uint iFilename, TRenameState Result)
{
	if (iFilename < m_Results.size())
	{
		// We only mark this result if it is more severe than the current
		// result
		if (Result < m_Results[iFilename])
		{
			SetFilenameStatus(iFilename, Result);
		}
	}
	else
	{
		COMPLAIN("Index out of bounds in MarkFilenameProblem()!");
	}
}

void FilenameList::SetFilenameStatus(uint iFilename, TRenameState Result)
{
	if (iFilename < m_Results.size())
	{
		// Update warning/error count, as we're about to change a result
		switch (m_Results[iFilename])
		{
			case RenameState_WarnNewFileExists:
			case RenameState_WarnFileNotFound:
			case RenameState_WarnNameClash:
			{
				m_cWarnings--;
				break;
			}

			case RenameState_ErrorFileNotFound:
			case RenameState_ErrorFileAlreadyExists:
			case RenameState_ErrorAccessDenied:
			case RenameState_ErrorFolderCreateFailed:
			case RenameState_ErrorUnknown:
			{
				m_cErrors--;
				break;
			}
		}

		m_Results[iFilename] = Result;

		switch (m_Results[iFilename])
		{
			case RenameState_WarnNewFileExists:
			case RenameState_WarnFileNotFound:
			case RenameState_WarnNameClash:
			{
				m_cWarnings++;
				break;
			}

			case RenameState_ErrorFileNotFound:
			case RenameState_ErrorFileAlreadyExists:
			case RenameState_ErrorAccessDenied:
			case RenameState_ErrorFolderCreateFailed:
			case RenameState_ErrorUnknown:
			{
				m_cErrors++;
				break;
			}
		}
	}
	else
	{
		COMPLAIN("Index out of bounds in SetFilenameStatus()!");
	}
}

int FilenameList::GetWarningCount()
{
	return m_cWarnings;
}

int FilenameList::GetErrorCount()
{
	return m_cErrors;
}

bool FilenameList::GetRenameState(uint iFilename, TRenameState* pResult)
{
	if (iFilename < m_Results.size())
	{
		*pResult = m_Results[iFilename];
		return true;
	}

	// Index out of bounds
	return false;
}
