//
// FilenameList.h
//

//
// This software is protected by the GNU General Public License,
// also known as the GPL.
//
// See the file COPYING.txt for details.
//

#ifndef inc_FilenameList_h
#define inc_FilenameList_h

#include "vgui\JobWorker.h"

#include "vos\FileSpec.h"

#include <vector>

#include <boost/regex.hpp>

#include "FileSpecs.h"

class ScanFilesWorker;
class FilterFilesWorker;
class GenerateFilenamesWorker;

//
// Indicates what happened when a rename was attempted on a file
//
typedef enum
{
	// Errors (after rename attempted)
	RenameState_ErrorInternal,
	RenameState_ErrorFileNotFound,
	RenameState_ErrorFileAlreadyExists,
	RenameState_ErrorAccessDenied,
	RenameState_ErrorFolderCreateFailed,
	RenameState_ErrorUnknown,

	// Warnings (before rename attempted)
	RenameState_WarnPatternFail,
	RenameState_WarnNameClash,
	RenameState_WarnBadMetaData,
	RenameState_WarnFileNotFound,
	RenameState_WarnNewFileExists,
	RenameState_WarnReadOnly,
	
	// Renamed ok
	RenameState_RenamedOK,

	// Rename not attempted, but looks like it will work
	RenameState_OK,
}
TRenameState;

typedef std::vector<TRenameState> RenameStates;

/**************************************************************************

>	class FilenameList

	Author:		Tim Browse
	Date:		30/4/2002

	Purpose:	Scan a folder and keep a list of files found in that 
				folder.  The list of files is optionally filtered by
				a standard wildcard spec, or a regular expression.
				
				The files also have new names generated for them, according
				to a renaming scheme.

**************************************************************************/

class FilenameList : public vgui::JobWorker
{
	public:
		typedef enum
		{
			Work_Rescan,
			Work_Refilter,
			Work_Preview,
			Work_Validate,
			Work_EnsureComplete,

			Work_Complete = Work_EnsureComplete	// Internal use only!
		}
		TWorkType;

	public:
		FilenameList();

		//
		// Setting up options for the work to be done...
		//

		// Enable progress display (when scanning disk for files)
		void EnableProgressDisplay(vgui::Window* pParent);
		void DisableProgressDisplay();

		// Change folder we're using
		void SetFolder(const vos::FileSpec& FolderName);
		vos::FileSpec GetFolder();

		// Filtering
		void DisableFiltering();
		void SetWildcardFilter(const vos::String& Filter, bool bIgnoreCase = true);
		void SetRegExpFilter(const vos::String& RegExp, bool bIgnoreCase = true);

		// Renaming
		void DisablePatternMatching();
		void SetPatternMatching(const vos::String& Pattern, bool bIgnoreCase = true);
		void SetRenameSpec(const vos::String& RenameSpec);

		//
		// Getting the processing done
		//
		// (ProgressWorker interface)
		//
		void PrepareToStartProcessing(TWorkType WorkType = Work_EnsureComplete);
		vgui::TJobStatus DoWork(vgui::JobController* pJobController,
							    int *pPercentComplete);
		void StopProcessing();

		//
		// Access the generated data
		//
		bool IsListFiltered();
		bool IsListValidated();
		bool IsListReady();
		uint GetFilenameCount();
		bool GetOriginalFilename(uint iFilename, vos::FileSpec* pfsFilename);
		bool GetNewFilename(uint iFilename, vos::FileSpec* pfsFilename);

		bool GetRenameState(uint iFilename, TRenameState* pResult);

		//
		// Generating data
		//
		void ClearErrors();
		void SetFilenameStatus(uint iFilename, TRenameState Result);
		void MarkFilenameProblem(uint iFilename, TRenameState Result);
		int GetWarningCount();
		int GetErrorCount();

	private:
		bool ScanFiles();
		bool FileMatchesFilter(const vos::FileSpec& fsFilename);
		
		void MoveToNextPhase();

		void InitialiseResults();

	private:
		// The folder we are scanning
		vos::FileSpec m_fsFolderName;

		//////////////////////////////////////////////////////////////////
		//
		// The lists of data we are generating...
		//
		//////////////////////////////////////////////////////////////////

		// The list of files (existing files, i.e. original filenames)
		FileSpecs m_Filenames;

		// An array of all filespec that match the current filter.
		// (Stored as indices into the m_Files vector)
		std::vector<uint> m_FilteredFilenames;

		// The new filenames with which we rename the files in m_Files.
		FileSpecs m_NewFilenames;

		// Status of each filename
		RenameStates m_Results;

		// Number of warnings encountering while scanning/previewing filenams
		int m_cWarnings;
		
		// Number of errors encountered while renaming files
		int m_cErrors;

		//////////////////////////////////////////////////////////////////
		//
		// Options to control processing
		//
		//////////////////////////////////////////////////////////////////

		//
		// Filtering options
		//
		// Do we use filtering, and if so, what is the filter pattern to use?
		bool m_bUseFilter;
		boost::RegEx m_Filter;

		// Filter represented by above object
		vos::String m_strFilter;
		bool m_bIgnoreCaseInFilter;

		//
		// Renaming options
		//
		// Do we match against a regexp pattern to extract tagged expressions
		// from the filenames, and if so, what is that pattern?
		bool m_bUsePatternMatching;
		vos::String m_Pattern;
		// The new name to use when renaming files - contains variables which
		// will be replaced to form new name.
		vos::String m_RenameSpec;

		//////////////////////////////////////////////////////////////////
		//
		// Current state of the processing (so we can pick up from where
		// we left off the next time ProcessFilenames() is called).
		//
		//////////////////////////////////////////////////////////////////

		// What needs to be done
		TWorkType m_WorkStatus;

		// Worker object currently active
		vgui::JobWorker* m_pWorker;
		
		// Allow worker object to access filter details etc
		friend class FilterFilesWorker;
		friend class GenerateFilenamesWorker;

		// Parent window for modal progress dialog
		vgui::Window* m_pParentWindow;


		// Friends
		friend class FilenameSorter;
};

#endif // !inc_FilenameList_h
