//
// FilenameTransformer.cpp
//

//
// This software is protected by the GNU General Public License,
// also known as the GPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "FilenameTransformer.h"

#include "MP3TagReader.h"


/**************************************************************************

>	FilenameTransformer::FilenameTransformer(const FilenameTransformSpec& Spec)

	Author:		Tim Browse
	Date:		29/4/2002
	Inputs:		Spec - details of how this object will transform filenames.

	Purpose:	Constructor

**************************************************************************/

FilenameTransformer::FilenameTransformer(const FilenameTransformSpec& Spec) : m_bUseMP3Tags(false)
{
	SetSpec(Spec);
}

/**************************************************************************

>	void FilenameTransformer::SetSpec(const FilenameTransformSpec& Spec)

	Author:		Tim Browse
	Date:		29/4/2002
	Inputs:		Spec - the transform spec to use

	Purpose:	Set the transform spec to use with this object.

**************************************************************************/

void FilenameTransformer::SetSpec(const FilenameTransformSpec& Spec)
{
	// Use this spec
	try
	{
		m_Spec = Spec;
	}
	catch(...)
	{
		// TODO: handle this!
	}
	
	// Ok, work out if we need to extract meta-data from mp3s
	m_bUseMP3Tags = m_Spec.m_NewNameSpec.Find("<mp3");
}

/**************************************************************************

>	vos::Error FilenameTransformer::TransformFilename(const vos::String& Filename,
													  vos::String* pNewFilename)


	Author:		Tim Browse
	Date:		29/4/2002
	Inputs:		Filename - the filename to transform
	Outputs:	pNewFilename - the transformed filename
	Returns:	TransformResult

	Purpose:	Process a particular filename, and transform it according
				to the FilenameTransformSpec passed to this object's
				constructor.

**************************************************************************/

TransformResult FilenameTransformer::TransformFilename(const vos::String& Filename,
													   vos::String* pNewFilename)
{
	// Reconstruct full pathname
	vos::FileSpec fsFullPathName(m_Spec.m_fsFolder);
	fsFullPathName.AppendFilename(Filename);

	// Ok, start with the pattern for the new name
	*pNewFilename = m_Spec.m_NewNameSpec;

	if (m_bUseMP3Tags)
	{
		// Ok, get the mp3 tags for this object.
		MP3TagReader Tags(fsFullPathName.GetString());
		vos::String Artist, Album, TrackName, TrackNumber;
		if (!Tags.GetInfo(&Artist, &Album, &TrackName, &TrackNumber))
		{
			// Failed to get meta-data from file
			return TransformResult_BadMetaData;
		}

		// Replace mp3 meta-data in this pattern
		pNewFilename->Replace("<mp3Artist>", Artist);
		pNewFilename->Replace("<mp3Album>",  Album);
		pNewFilename->Replace("<mp3TrackName>", TrackName);
		pNewFilename->Replace("<mp3TrackNumber>", TrackNumber);
	}

	// Get the tagged expressions in the string
	boost::match_results<const char*> TaggedExpressions;
	if (!boost::regex_match(Filename.c_str(), TaggedExpressions, m_Spec.m_RegExp))
	{
		// Pattern did not match
		return TransformResult_PatternMatchFail;
	}

	// Do a s&r on the pattern for this data.
	int iTagIndex = 0;

	while (iTagIndex < TaggedExpressions.length())
	{
		vos::String strExpr;
		strExpr.Printf("%%%d", iTagIndex);

		boost::sub_match<char const*> SubExpr = TaggedExpressions[iTagIndex];

		std::string S = std::string(SubExpr.first, SubExpr.second);

		pNewFilename->Replace(strExpr.c_str(), S.c_str());

		iTagIndex++;
	}

	return TransformResult_OK;
}

vos::FileSpec FilenameTransformer::GetFolder()
{
	return m_Spec.m_fsFolder;
}
