//
// FilenameValidateWorker.cpp
//

//
// This software is protected by the GNU General Public License,
// also known as the GPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "FilenameValidateWorker.h"

#include "vos\MonotonicTime.h"
#include "vos\Trace.h"

#include "FilenameList.h"
#include "FilenameSorter.h"

#include <algorithm>

FilenameValidateWorker::FilenameValidateWorker(FilenameList* pFilenameList)

	  : m_pFilenameList(pFilenameList),
		m_cFiles(0),
		m_iCurrentFile(0)
{
	// Find out how many files we are to process.
	m_cFiles = m_pFilenameList->GetFilenameCount();
	m_iCurrentFile = 0;
}

vgui::TJobStatus FilenameValidateWorker::DoWork(vgui::JobController* pJobController,
												int* pPercentComplete)
{
	if (m_cFiles == 0)
	{
		// No files found/matched, so we're done
		return vgui::Job_Finished;
	}

	if (m_iCurrentFile < m_cFiles)
	{
		// Do file existence/read-only checks
		DoFileChecks(pJobController, pPercentComplete);

		// Adjust percentage to be half the work
		*pPercentComplete /= 2;

		// We're not done yet - still need to check for duplicates.
		return vgui::Job_NotFinished;
	}

	// Ok, time to sort the list of filenames
	m_SortedIDs.resize(m_cFiles);

	// Initialise to match new filename array, with no sorting
	int iFile;
	for (iFile = 0; iFile < m_cFiles; iFile++)
	{
		m_SortedIDs[iFile] = iFile;
	}

	// Ok, now sort the array of IDs, based on the filenames they reference.
	std::sort(m_SortedIDs.begin(), m_SortedIDs.end(), 
	 		  FilenameSorter(m_pFilenameList));

	// Check for dupes
	vos::FileSpec fsName1, fsName2;
	for (iFile = 0; iFile < (m_cFiles - 1); iFile++)
	{
		m_pFilenameList->GetNewFilename(iFile, &fsName1);
		m_pFilenameList->GetNewFilename(iFile+1, &fsName2);

		if (fsName1 == fsName2)
		{
			// Duplicate found - mark both with clash warning
			m_pFilenameList->MarkFilenameProblem(iFile, RenameState_WarnNameClash);
			m_pFilenameList->MarkFilenameProblem(iFile+1, RenameState_WarnNameClash);
		}
	}

	return vgui::Job_Finished;
}

vgui::TJobStatus FilenameValidateWorker::DoFileChecks(vgui::JobController* pJobController,
													  int* pPercentComplete)
{
	int cFilesToProcess = 25;

	vos::FileSpec fsFolder(m_pFilenameList->GetFolder());

	vos::FileSpec fsLeafName;
	vos::FileSpec fsPathName;

	TRenameState State;

	while ((cFilesToProcess > 0) && (m_iCurrentFile < m_cFiles))
	{
		// Get current state, and modify if we find warnings/errors
		m_pFilenameList->GetRenameState(m_iCurrentFile, &State);

		// Check that the source file exists...
		if (m_pFilenameList->GetOriginalFilename(m_iCurrentFile, &fsLeafName))
		{
			fsPathName = fsFolder;
			fsPathName.AppendRelativeFileSpec(fsLeafName);

			if (fsPathName.GetFileType() != vos::FileType_File)
			{
				State = RenameState_WarnFileNotFound;
			}
			else
			{
				// It does exist - is it read-only?
				if (fsPathName.IsReadOnly())
				{
					// NB. this warning is not displayed if the user
					//	   has decided to always rename read-only files.
					State = RenameState_WarnReadOnly;
				}
			}
		}
		else
		{
			// Problem getting the original filename...
			State = RenameState_ErrorUnknown;
		}

		// Ok, validate the new name...
		if (m_pFilenameList->GetNewFilename(m_iCurrentFile, &fsLeafName))
		{
			// Ok, does the new name exist yet?
			fsPathName = fsFolder;
			fsPathName.AppendRelativeFileSpec(fsLeafName);

			if (fsPathName.GetFileType() != vos::FileType_NotFound)
			{
				// Yes, so we can't rename to that
				State = RenameState_WarnNewFileExists;
			}
		}
		else
		{
			// Problem getting the new filename...
			State = RenameState_ErrorUnknown;
		}

		// Ok, set the state of the file based on this info
		m_pFilenameList->MarkFilenameProblem(m_iCurrentFile, State);

		// Try next file
		m_iCurrentFile++;
		cFilesToProcess--;
	}

	// Tell progress dialog how we're doing.
	if (m_cFiles > 0)
	{
		*pPercentComplete = (100 * m_iCurrentFile) / m_cFiles;
	}
	else
	{
		*pPercentComplete = 100;
	}

	// And tell them if we've finished yet
	if (m_iCurrentFile < m_cFiles)
	{
		// Still more filenames to show
		return vgui::Job_NotFinished;
	}
	else
	{
		// All done
		return vgui::Job_Finished;
	}
}

void FilenameValidateWorker::OnCancel(vgui::JobController* pJobController)
{
	// Nothing to do really...the display will get updated in the background
	// soon.
}

