//
// GenerateFilenamesWorker.cpp
//

//
// This software is protected by the GNU General Public License,
// also known as the GPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "GenerateFilenamesWorker.h"

#include "FilenameList.h"

/**************************************************************************

>	GenerateFilenamesWorker::GenerateFilenamesWorker(FilenameList* pFilenameList)

	Author:		Tim Browse
	Date:		30/4/2002
	Inputs:		pFilenameList - the filename list object to use.

	Purpose:	Constructor

**************************************************************************/

GenerateFilenamesWorker::GenerateFilenamesWorker(FilenameList* pFilenameList) : m_pFilenameList(pFilenameList),
																				m_bFinished(false),
																				m_cFiles(0),
																				m_iCurrentFile(0)
{
	// Clear the new filenames array before we start
	m_pFilenameList->m_NewFilenames.clear();

	// Find out how many files we have
	m_cFiles = m_pFilenameList->m_FilteredFilenames.size();

	// Set up filename transform
	FilenameTransformSpec Spec;
	Spec.SetSourcePattern(m_pFilenameList->m_Pattern);
	Spec.m_fsFolder = m_pFilenameList->GetFolder();
	Spec.m_NewNameSpec = m_pFilenameList->m_RenameSpec;

	m_Transformer.SetSpec(Spec);
}

/**************************************************************************

>	vgui::TJobStatus GenerateFilenamesWorker::DoWork(vgui::JobController* pJobController,
																int* pPercentComplete)

	Author:		Tim Browse
	Date:		30/4/2002
	Outputs:	pPercentComplete - the current completion status.
	Returns:	JobWorker_Finished - work has complete
				JobWorker_NotFinished - work is not finished, and the
											 worker object is expecting this
											 function to be called again.

	Purpose:	Generate a batch of new filenames

**************************************************************************/

vgui::TJobStatus GenerateFilenamesWorker::DoWork(vgui::JobController* pJobController,
															int* pPercentComplete)
{
	m_bFinished = false;

	int cFilenamesToGenerate = 50;

	vos::String NewFilename;

	while ((cFilenamesToGenerate > 0) && (m_iCurrentFile < m_cFiles))
	{
		// Generate new name for this file.
		uint iFilteredFile = m_pFilenameList->m_FilteredFilenames[m_iCurrentFile];

		vos::String OriginalFilename(m_pFilenameList->m_Filenames[iFilteredFile].GetString());

		switch (m_Transformer.TransformFilename(OriginalFilename, &NewFilename))
		{
			case TransformResult_BadMetaData:
			{
				m_pFilenameList->MarkFilenameProblem(m_iCurrentFile, 
												     RenameState_WarnBadMetaData);
				break;
			}

			case TransformResult_PatternMatchFail:
			{
				m_pFilenameList->MarkFilenameProblem(m_iCurrentFile, 
												     RenameState_WarnPatternFail);
				break;
			}
		}

		// Add to our list
		m_pFilenameList->m_NewFilenames.push_back(NewFilename);

		// Try next file
		m_iCurrentFile++;
		cFilenamesToGenerate--;
	}

	// Tell caller how far we've got
	if (m_cFiles > 0)
		*pPercentComplete = (100 * m_iCurrentFile) / m_cFiles;
	else
		*pPercentComplete = 100;

	// Tell caller if we're done yet
	if (m_iCurrentFile == m_cFiles)
	{
		m_bFinished = true;
		return vgui::Job_Finished;
	}
	else
	{
		return vgui::Job_NotFinished;
	}
}

/**************************************************************************

>	void GenerateFilenamesWorker::OnCancel(vgui::JobController* pJobController)

	Author:		Tim Browse
	Date:		30/4/2002
	Inputs:		pProgressDialog - the progress dialog being used.

	Purpose:	Handle the user cancelling the request

**************************************************************************/

void GenerateFilenamesWorker::OnCancel(vgui::JobController* pJobController)
{
	// Clear out the filter list
	m_pFilenameList->m_NewFilenames.clear();
}
