//
// RenameUndoManager.cpp
//

//
// This software is protected by the GNU General Public License,
// also known as the GPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "RenameUndoManager.h"

#include "vos\Errors.h"
#include "vos\MonotonicTime.h"

#include "FilenameList.h"

/**************************************************************************

>	RenameUndoManager::RenameUndoManager()

	Author:		Tim Browse
	Date:		4/5/2002

	Purpose:	Constructor

**************************************************************************/

RenameUndoManager::RenameUndoManager() : m_cFilesToRename(0),
										 m_cFoldersToDelete(0),
										 m_iCurrentFile(0),
										 m_iCurrentFolder(0),
										 m_pFilenameList(NULL)
{
	// TCILB
}

/**************************************************************************

>	void RenameUndoManager::LinkToFilenameList(FilenameList* pFilenameList)

	Author:		Tim Browse
	Date:		12/5/2002
	Inputs:		pFilenameList - the new list object to use

	Purpose:	Link this undo manager to the specified FilenameList
				object.  This means the file indices passed to 
				NotifyFileRenamed() can be used to extract file info
				from this FilenameList.

**************************************************************************/

void RenameUndoManager::LinkToFilenameList(FilenameList* pFilenameList)
{
	m_pFilenameList = pFilenameList;
}

/**************************************************************************

>	void RenameUndoManager::ClearHistory()

	Author:		Tim Browse
	Date:		3/5/2002

	Purpose:	Clear out the undo history, ready for a new set of data to
				be added via the IFileRenameMonitor interface.

**************************************************************************/

void RenameUndoManager::ClearHistory()
{
	m_OriginalFilenames.clear();
	m_NewFilenames.clear();
	m_Indices.clear();
	m_FoldersCreated.clear();
}

/**************************************************************************

>	void RenameUndoManager::NotifyFileRenamed(uint iFileIndex,
											  const vos::FileSpec& fsOldName,
											  const vos::FileSpec& fsNewName)


	Author:		Tim Browse
	Date:		3/5/2002
	Inputs:		iFileIndex - index of the file being renamed (used to
							 communicate with the FilenameList object)
				fsOldName - the original name of the file.
				fsNewName - the name the file has now.

	Purpose:	Record the details of the file being renamed, and its new
				name, so we can undo it later if required.

**************************************************************************/

void RenameUndoManager::NotifyFileRenamed(uint iFileIndex,
										  const vos::FileSpec& fsOldName,
										  const vos::FileSpec& fsNewName)
{
	m_OriginalFilenames.push_back(fsOldName);
	m_NewFilenames.push_back(fsNewName);
	m_Indices.push_back(iFileIndex);
}

/**************************************************************************

>	void RenameUndoManager::NotifyFolderCreated(const vos::FileSpec& fsNewFolder)

	Author:		Tim Browse
	Date:		3/5/2002
	Inputs:		fsNewFolder - name of the new folder that has been created.

	Purpose:	Record details of a folder that has been created as part
				of a rename operation.

**************************************************************************/

void RenameUndoManager::NotifyFolderCreated(const vos::FileSpec& fsNewFolder)
{
	m_FoldersCreated.push_back(fsNewFolder);
}

/**************************************************************************

>	void RenameUndoManager::NotifyFolderDeleted(const vos::FileSpec& fsDeletedFolder)

	Author:		Tim Browse
	Date:		3/5/2002
	Inputs:		fsDeletedFolder - name of the folder that has been deleted.

	Purpose:	Record details of a folder that has been deleted due to an
				error while renaming a file.

**************************************************************************/

void RenameUndoManager::NotifyFolderDeleted(const vos::FileSpec& fsDeletedFolder)
{
	// Find this folder and remove it - we search backwards for speed
	// (not that it's going to make much difference)
	FileSpecs::iterator itFolder = m_FoldersCreated.end();

	while (itFolder != m_FoldersCreated.begin())
	{
		if (*itFolder == fsDeletedFolder)
		{
			// Found folder - remove from our list
			m_FoldersCreated.erase(itFolder);
			return;
		}
		
		// Try previous folder
		--itFolder;
	}
}

/**************************************************************************

>	vgui::TJobStatus RenameUndoManager::DoWork(vgui::JobController* pJobController,
													  int* pPercentComplete)


	Author:		Tim Browse
	Date:		3/5/2002
	Inputs:		pJobController - the controller in charge of this job
	Outputs:	pPercentComplete - how far through the job we are
	Returns:	See JobWorker_Timesliced

	Purpose:	Undo the file rename operation that this object holds the	
				undo records for.

**************************************************************************/

vgui::TJobStatus RenameUndoManager::DoWork(vgui::JobController* pJobController,
										   int* pPercentComplete)
{
	//
	// Handle file renaming
	//
	if ((m_cFilesToRename > 0) && (m_iCurrentFile < m_cFilesToRename))
	{
		int cFilesToRenameNow = 25;

		while ((cFilesToRenameNow > 0) &&
			   (m_iCurrentFile < m_cFilesToRename))
		{
			// Ok, time to rename some files - we rename from new name back
			// to original.
			if (m_NewFilenames[m_iCurrentFile].Rename(m_OriginalFilenames[m_iCurrentFile]))
			{
				// File unrenamed ok
				if (m_pFilenameList != NULL)
				{
					m_pFilenameList->SetFilenameStatus(m_Indices[m_iCurrentFile], 
													   RenameState_OK);
				}
			}
			else
			{
				// Rename failed...
				// TODO: Record the error so we can show the user.
				ERROR_HANDLED();
			}

			cFilesToRenameNow--;
			m_iCurrentFile++;
		}

		// Finished for now
		int cTotalObjects = m_cFilesToRename + m_cFoldersToDelete;
		if (cTotalObjects > 0)
		{
			*pPercentComplete = (m_iCurrentFile * 100) / cTotalObjects;
		}

		// Have we finished completely?
		if ((m_iCurrentFile < m_cFilesToRename) || (m_cFoldersToDelete > 0))
		{
			// No - we still have work to do
			return vgui::Job_NotFinished;
		}
		else
		{
			// Nothing left to do
			return vgui::Job_Finished;
		}
	}

	// 
	// If all files have been restored, now we remove the folders.
	//
	// NB. Folders are removed in reverse order, for reasons which will become
	//	   obvious if you think about it for a couple of seconds.
	//
	if ((m_cFoldersToDelete> 0) && (m_iCurrentFolder >= 0))
	{
		int cFoldersToDeleteNow = 25;

		while ((cFoldersToDeleteNow > 0) && (m_iCurrentFolder >= 0))
		{
			// Ok, time to rename some files - we rename from new name back
			// to original.
			if (!m_FoldersCreated[m_iCurrentFolder].Delete())
			{
				// Folder removal failed...
				// TODO: Record the error so we can show the user.
				ERROR_HANDLED();
			}

			cFoldersToDeleteNow--;
			m_iCurrentFolder--;
		}

		// Finished for now
		int cTotalObjects = m_cFilesToRename + m_cFoldersToDelete;
		if (cTotalObjects > 0)
		{
			int cFoldersDeleted = m_cFoldersToDelete - (m_iCurrentFolder+1);
			*pPercentComplete = ((m_cFilesToRename + cFoldersDeleted) * 100) / cTotalObjects;
		}

		// Have we finished completely?
		if (m_iCurrentFolder >= 0)
		{
			// No - we still have work to do
			return vgui::Job_NotFinished;
		}
		else
		{
			// Nothing left to do
			return vgui::Job_Finished;
		}
	}

	// Ok, we've renamed all the files, and removed all the folders,
	// so we're done.
	*pPercentComplete = 100;
	return vgui::Job_Finished;
}


/**************************************************************************

>	bool RenameUndoManager::PossibleToUndo()

	Author:		Tim Browse
	Date:		3/5/2002
	Returns:	true => undo records exist;
				false => no undo information available

	Purpose:	Find out if it's possible to undo anything, i.e. if this
				object has any undo records to execute.

**************************************************************************/

bool RenameUndoManager::PossibleToUndo()
{
	// If we have any file/folder info, then we can undo
	return (!m_OriginalFilenames.empty() || !m_FoldersCreated.empty());
}

/**************************************************************************

>	void RenameUndoManager::PrepareToUndo()

	Author:		Tim Browse
	Date:		4/5/2002

	Purpose:	Called before DoWork() is called, so the object can prepare
				to do its job.

**************************************************************************/

void RenameUndoManager::PrepareToUndo()
{
	m_cFilesToRename = m_OriginalFilenames.size();
	m_iCurrentFile = 0;

	// We delete folders in reverse order
	m_cFoldersToDelete = m_FoldersCreated.size();
	m_iCurrentFolder = m_cFoldersToDelete - 1;
}
