//
// ScanFilesWorker.cpp
//

//
// This software is protected by the GNU General Public License,
// also known as the GPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "ScanFilesWorker.h"

/**************************************************************************

>	ScanFilesWorker::ScanFilesWorker(const vos::FileSpec& fsFolderToScan,
									 FileSpecs* pFileSpecs)

	Author:		Tim Browse
	Date:		30/4/2002
	Inputs:		fsFolderToScan - the folder to scan for files
				pFileSpecs - the filespec container to fill with details
							 of the files that are scanned.

	Purpose:	Constructor - specify which folder to scan for files.

**************************************************************************/

ScanFilesWorker::ScanFilesWorker(const vos::FileSpec& fsFolderToScan,
								 FileSpecs* pFileSpecs) : m_fsFolderToScan(fsFolderToScan),
														  m_pFileSpecs(pFileSpecs),
														  m_bReadyToScan(false),
														  m_bFinished(false),
														  m_bFolderExists(false)
{
	// Clear the file list before we start
	m_pFileSpecs->clear();

	// Check the folder exists first
	if (m_fsFolderToScan.GetFileType() == vos::FileType_Folder)
	{
		m_bFolderExists = true;
	}
}

/**************************************************************************

>	vgui::TJobStatus ScanFilesWorker::DoWork(vgui::JobController* pJobController,
											 int* pPercentComplete)

	Author:		Tim Browse
	Date:		30/4/2002
	Outputs:	pPercentComplete - the current completion status.
	Returns:	Job_Finished - work has complete
				Job_NotFinished - work is not finished, and the worker object 
								  is expecting this function to be called again.

	Purpose:	Scan a batch of files.

**************************************************************************/

vgui::TJobStatus ScanFilesWorker::DoWork(vgui::JobController* pJobController,
										 int* pPercentComplete)
{
	if (!m_bFolderExists)
	{
		// If folder to scan does not exist, we're done.
		*pPercentComplete = 0;
		return vgui::Job_Finished;
	}

	m_bFinished = false;

	// If this is the first time we're called, we need to initialise
	// the scanning object
	if (!m_bReadyToScan)
		PrepareToScan();

	// Scan a chunk of files
	vos::FileSpec FoundSpec;
	int cFilesToScan = 50;

	while ((cFilesToScan > 0) && m_MatchingFiles.GetNextFile(&FoundSpec))
	{
		if (FoundSpec.GetFileType() == vos::FileType_File)
		{
			// Get the leafname of this file
			vos::FileSpec fsLeafname(FoundSpec.GetLeafname());
			m_pFileSpecs->push_back(fsLeafname);
		}

		cFilesToScan--;
	}

	// Tell caller if we're done yet
	if (cFilesToScan == 0)
	{
		return vgui::Job_NotFinished;
	}
	else
	{
		m_bFinished = true;
		return vgui::Job_Finished;
	}
}

/**************************************************************************

>	void ScanFilesWorker::PrepareToScan()

	Author:		Tim Browse
	Date:		30/4/2002

	Purpose:	Get ready to scan for files - intialises the FileMatches
				object to scan the desired folder.

**************************************************************************/

void ScanFilesWorker::PrepareToScan()
{
	// Initialise FileMatches object
	vos::FileSpec fsMatchSpec(m_fsFolderToScan);
	fsMatchSpec.AppendFilename("*.*");

	m_MatchingFiles.StartScan(fsMatchSpec);

	// All done
	m_bReadyToScan = true;
}

/**************************************************************************

>	void ScanFilesWorker::OnCancel(vgui::JobController* pJobController)

	Author:		Tim Browse
	Date:		30/4/2002
	Inputs:		pProgressDialog - the progress dialog being used.

	Purpose:	Handle the user cancelling the scan request

**************************************************************************/

void ScanFilesWorker::OnCancel(vgui::JobController* pJobController)
{
	// Clean out the list of files
	m_pFileSpecs->clear();
}

